#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Reader for bioformat supported files."""

import math
import os
from contextlib import contextmanager
from functools import cached_property, lru_cache
from pathlib import Path
from queue import Empty, SimpleQueue
from tempfile import TemporaryDirectory
from threading import Lock
from typing import Dict, Generator, List, Optional, Tuple, Type, Union

import jpype.imports  # Needed for loci import to work # noqa
import numpy as np
import ome_types
import scyjava
from jpype.types import JArray
from wsidicom.geometry import Region, Size, SizeMm

"""
Set version of bioformats jar to use with the environmental variable
"BIOFORMATS_VERSION". Note the bioformats jar either has a BSD or GPL-2
license.
"""
bioformats_version = os.getenv("BIOFORMATS_VERSION", "bsd:8.3.0")
scyjava.config.endpoints.append(f"ome:formats-{bioformats_version}")

if not scyjava.jvm_started():
    scyjava.start_jvm()

"""
Load bioformats modules using jpype and scyjava
"""
from loci.common.services import ServiceFactory  # type: ignore # noqa
from loci.formats import ImageReader, Memoizer  # type: ignore # noqa
from loci.formats.services import OMEXMLService  # type: ignore # noqa


class ReaderPool:
    def __init__(
        self,
        filepath: Path,
        max_readers: Optional[int] = None,
        cache_path: Optional[Union[Path, str]] = None,
    ):
        self._filepath = filepath
        if max_readers is None:
            cpu_count = os.cpu_count()
            if cpu_count is not None:
                max_readers = cpu_count
            else:
                max_readers = 1
        self._max_readers = max_readers
        self._current_count = 0
        if cache_path is None:
            self._tempdir = TemporaryDirectory()
            self._cache_path = Path(self._tempdir.name)
        else:
            self._tempdir = None
            self._cache_path = Path(cache_path)
        self._queue: SimpleQueue[Memoizer] = SimpleQueue()
        self._lock = Lock()

    @property
    def filepath(self) -> Path:
        return self._filepath

    @contextmanager
    def get_reader(self) -> Generator[Memoizer, None, None]:
        """Return a reader. Should be used as a context manager. Will block
        if no reader is available."""
        reader = self._get_reader()
        try:
            yield reader
        finally:
            self._return_reader(reader)

    @property
    def _is_full(self) -> bool:
        return self._current_count == self._max_readers

    @property
    def _is_empty(self) -> bool:
        return self._current_count == 0

    def _increment(self):
        if self._current_count < self._max_readers:
            self._current_count += 1
            return
        raise ValueError("Current count should never be higher than max readers.")

    def _decrement(self):
        if self._current_count > 0:
            self._current_count -= 1
            return
        raise ValueError("Current count should never be lower than 0.")

    def _get_reader(self) -> Memoizer:
        """Return a reader with no wait if one is available. Else if current
        reader count is less than maximum reader count return a new reader.
        Otherwise wait for an available readern."""
        try:
            return self._queue.get_nowait()
        except Empty:
            pass
        with self._lock:
            if not self._is_full:
                self._increment()
                return self._create_new_reader()

        return self._queue.get(block=True, timeout=None)

    def _create_new_reader(self) -> Memoizer:
        """Return a new reader."""
        # Create a reader using Memoizer to load file faster
        # See https://docs.openmicroscopy.org/bio-formats/6.11.0/developers/matlab-dev.html#reader-performance
        reader = Memoizer(ImageReader(), 0, self._cache_path)
        reader.setFlattenedResolutions(False)
        reader.setId(str(self._filepath))
        return reader

    def _return_reader(self, reader: Memoizer) -> None:
        """Release a used reader."""
        self._queue.put(reader)

    def close(self) -> None:
        """Close all readers and clean up cache directory if a temporary directory."""
        with self._lock:
            while not self._is_empty:
                reader = self._queue.get()
                reader.close()
                self._decrement()
            if self._tempdir is not None:
                self._tempdir.cleanup()


class BioformatsReader:
    def __init__(
        self,
        filepath: Path,
        max_readers: Optional[int] = None,
        cache_path: Optional[Union[Path, str]] = None,
    ):
        """Reader for image data and metadata from file using Bio-Formats api.

        Parameters
        ----------
        filepath: Path
            Path to file to open.
        max_readers: int
            Maximum number of readers to use. If not specified the number
            of cpus will be used.
        cache_path: Optional[Union[Path, str]] = None
            Path to store cache file quicker opening of new readers. If None
            a temporary directory will be used.
        """
        self._reader_pool = ReaderPool(filepath, max_readers, cache_path)

    @staticmethod
    def is_supported(filepath: Path) -> bool:
        try:
            reader = ImageReader()
            reader.setId(str(filepath))
            reader.close()
        except Exception:
            return False
        return True

    @property
    def filepath(self) -> Path:
        """Return filepath of the opened file."""
        return self._reader_pool.filepath

    @cached_property
    def metadata(self) -> ome_types.OME:
        """Return parsed metadata."""
        metadata = self._read_metadata()
        return ome_types.from_xml(str(metadata), parser="lxml")

    @property
    def images_count(self) -> int:
        """Return number of images in file."""
        return len(self.metadata.images)

    def image_name(self, image_index: int) -> Optional[str]:
        """Return name of image."""
        return self.metadata.images[image_index].name

    @lru_cache
    def dtype(self, image_index: int) -> np.dtype:
        """Return the numpy datatype for image in file."""
        NUMPY_DATA_TYPES: Dict[str, Type] = {
            "bit": np.bool_,
            "double-complex": np.cdouble,
            "complex": np.csingle,
            "double": np.float64,
            "float": np.float32,
            "int16": np.int16,
            "int32": np.int32,
            "int8": np.int8,
            "uint16": np.uint16,
            "uint32": np.uint32,
            "uint8": np.uint8,
        }
        if self.metadata.images[image_index].pixels.big_endian:
            byte_order = ">"
        else:
            byte_order = "<"
        data_type = self.metadata.images[image_index].pixels.type.value
        try:
            numpy_data_type = np.dtype(NUMPY_DATA_TYPES[data_type])
        except KeyError:
            raise ValueError(f"Unkown data type {data_type}")
        return numpy_data_type.newbyteorder(byte_order)

    @lru_cache
    def samples_per_pixel(self, image_index: int) -> int:
        """Return the samples per pixel for image in file."""
        pixels = self.metadata.images[image_index].pixels
        samples_per_pixel = pixels.channels[0].samples_per_pixel
        assert samples_per_pixel is not None
        return int(samples_per_pixel)

    @lru_cache
    def size(self, image_index: int, resolution_index: int = 0) -> Size:
        """Return the image size for image in file."""
        with self._reader_pool.get_reader() as reader:
            reader.setSeries(image_index)
            reader.setResolution(resolution_index)
            width = reader.getSizeX()
            height = reader.getSizeY()
            return Size(int(width), int(height))

    @lru_cache
    def pixel_spacing(
        self, image_index: int, resolution_index: int = 0
    ) -> Optional[SizeMm]:
        """Return the size of the pixels in mm/pixel for image in file."""
        pixels = self.metadata.images[image_index].pixels
        if pixels.physical_size_x is None or pixels.physical_size_y is None:
            return None
        scale = self._resolution_scales(image_index)[resolution_index]
        return (
            SizeMm(float(pixels.physical_size_x), float(pixels.physical_size_y))
            * scale
            / 1000
        )

    @lru_cache
    def is_interleaved(self, image_index: int) -> bool:
        """Return true if image data is interleaved."""
        interleaved = self.metadata.images[image_index].pixels.interleaved
        assert interleaved is not None
        return interleaved

    @lru_cache
    def pyramid_levels(self, image_index: int) -> Dict[Tuple[int, float, str], int]:
        """Return dictionary of dyadic scaling, focal plane, and optical path as key
        and resolution index as value for resolutions in image.

        Focal planes and optical paths are not implemented.
        """
        TOLERANCE = 1e-2
        float_pyramid_levels = (
            math.log2(scale) for scale in self._resolution_scales(image_index)
        )
        return {
            (round(float_level), 0.0, "0"): resolution_index
            for resolution_index, float_level in enumerate(float_pyramid_levels)
            if math.isclose(float_level, round(float_level), abs_tol=TOLERANCE)
        }

    @lru_cache
    def _resolution_scales(self, image_index: int) -> List[float]:
        """Return resolution scales for image as resolution width divided by image width."""
        with self._reader_pool.get_reader() as reader:
            reader.setSeries(image_index)
            reader.setResolution(0)
            base_width = reader.getSizeX()
            return [
                self._get_resolution_scale(reader, resolution_index, base_width)
                for resolution_index in range(reader.getResolutionCount())
            ]

    @staticmethod
    def _get_resolution_scale(
        reader: Memoizer, resolution_index: int, base_width: int
    ) -> float:
        """Return resolution scale for resolution as rounded int of resolution width
        divided by base widht."""
        reader.setResolution(resolution_index)
        width = reader.getSizeX()
        return base_width / width

    @contextmanager
    def read_image(
        self,
        image_index: int,
        resolution_index: int,
        region: Region,
        output_size: Optional[Size] = None,
        index: int = 0,
    ) -> Generator[np.ndarray, None, None]:
        """Read image data from file. Preferably used as a context manager.
        Data is returned as a memoryview that should be released after use.

        Parameters
        ----------
        image_index: int
            The image to read data from.
        region: Region
            The region to read data from.
        output_size: Optional[Size] = None
            Optional size to resize image data to. Must be equal to or larger than
            region size.
        index: int
            The index in image to read data from.

        Returns
        ----------
            Generator[np.ndarray, None, None]

        """
        try:
            raw_data = memoryview(
                self._read(
                    image_index,
                    resolution_index,
                    index,
                    region.start.x,
                    region.start.y,
                    region.size.width,
                    region.size.height,
                )
            )
        except Exception as exception:
            raise Exception(
                f"Failed to read image data from image {image_index}, resolution {resolution_index}, index {index} at region {region.box}."
            ) from exception
        try:
            data: np.ndarray = np.frombuffer(raw_data, self.dtype(image_index))
            if self.is_interleaved(image_index):
                data = data.reshape(
                    region.size.height,
                    region.size.width,
                    self.samples_per_pixel(image_index),
                )
            else:
                data = data.reshape(
                    self.samples_per_pixel(image_index),
                    region.size.height,
                    region.size.width,
                )
                data = np.moveaxis(data, 0, 2).copy()
            if output_size is not None and data.shape[0:2] != output_size.to_tuple():
                # Pad with zeros to get requsted output size.
                if not output_size.all_greater_than_or_equal(region.size):
                    raise ValueError(
                        "Output size should be equal to or larger than region size."
                    )
                padding_width = output_size.width - data.shape[0]
                padding_height = output_size.height - data.shape[1]
                data = np.pad(data, ((0, padding_width), (0, padding_height), (0, 0)))
            yield data
        finally:
            raw_data.release()

    def close(self) -> None:
        """Close reader pool."""
        self._reader_pool.close()

    def _read(
        self,
        image_index: int,
        resolution_index: int,
        index: int,
        start_x: int,
        start_y: int,
        end_x: int,
        end_y: int,
    ) -> JArray:
        """Read image data from file."""
        with self._reader_pool.get_reader() as reader:
            reader.setSeries(image_index)
            reader.setResolution(resolution_index)
            return reader.openBytes(index, start_x, start_y, end_x, end_y)

    def _read_metadata(self) -> str:
        """Read metadata from file."""
        with self._reader_pool.get_reader() as reader:
            metadata_store = reader.getMetadataStore()
            return str(metadata_store.dumpXML())



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Source for reading opentile compatible file."""

from functools import cached_property
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

from opentile import OpenTile
from pydicom import Dataset
from wsidicom.codec import Encoder
from wsidicom.geometry import Size, SizeMm
from wsidicom.metadata.wsi import WsiMetadata

from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.image_data import DicomizerImageData
from wsidicomizer.metadata import MetadataPostProcessor
from wsidicomizer.sources.opentile.opentile_image_data import (
    OpenTileAssociatedImageData,
    OpenTileLevelImageData,
)
from wsidicomizer.sources.opentile.opentile_metadata import OpenTileMetadata


class OpenTileSource(DicomizerSource):
    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: int = 512,
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        include_confidential: bool = True,
        metadata_post_processor: Optional[Union[Dataset, MetadataPostProcessor]] = None,
        force_transcoding: bool = False,
    ) -> None:
        """Create a new OpenTileSource.

        Parameters
        ----------
        filepath: Path
            Path to the file.
        encoder: Encoder
            Encoder to use. Pyramid is always re-encoded using the encoder.
        tile_size: int = 512,
            Preferred tile size to use, if not enforced by file.
        metadata: Optional[WsiMetadata] = None
            User-specified metadata that will overload metadata from source image file.
        default_metadata: Optional[WsiMetadata] = None
            User-specified metadata that will be used as default values.
        include_confidential: bool = True
            Include confidential metadata.
        metadata_post_processor: Optional[Union[Dataset, MetadataPostProcessor]] = None
            Optional metadata post processing by update from dataset or callback.
        force_transcoding: bool = False
            If to force transcoding images.
        """
        self._tiler = OpenTile.open(filepath, tile_size)
        self._base_metadata = OpenTileMetadata(
            self._tiler.metadata,
            self.has_label,
            self.has_overview,
            self._tiler.icc_profile,
        )

        self._force_transcoding = force_transcoding
        super().__init__(
            filepath,
            encoder,
            tile_size,
            metadata,
            default_metadata,
            include_confidential,
            metadata_post_processor,
        )

    def close(self):
        self._tiler.close()

    @property
    def base_metadata(self) -> OpenTileMetadata:
        return self._base_metadata

    @property
    def pyramid_levels(self) -> Dict[Tuple[int, float, str], int]:
        return {
            (level.pyramid_index, level.focal_plane, level.optical_path): index
            for index, level in enumerate(self._tiler.levels)
        }

    @property
    def has_label(self) -> bool:
        return len(self._tiler.labels) > 0

    @property
    def has_overview(self) -> bool:
        return len(self._tiler.overviews) > 0

    @staticmethod
    def is_supported(filepath: Path) -> bool:
        """Return True if file in filepath is supported by OpenTile."""
        return OpenTile.detect_format(filepath) is not None

    def _create_level_image_data(self, level_index: int) -> DicomizerImageData:
        return OpenTileLevelImageData(
            self._tiler.levels[level_index],
            self.base_metadata.pyramid.image,
            self.metadata.pyramid.image,
            self._encoder,
            self._volume_imaged_size,
            self._force_transcoding,
        )

    def _create_label_image_data(self) -> Optional[DicomizerImageData]:
        if not self.has_label:
            return None
        return OpenTileAssociatedImageData(
            self._tiler.labels[0], self._encoder, self._force_transcoding
        )

    def _create_overview_image_data(self) -> Optional[DicomizerImageData]:
        if not self.has_overview:
            return None
        return OpenTileAssociatedImageData(
            self._tiler.overviews[0], self._encoder, self._force_transcoding
        )

    def _create_thumbnail_image_data(self) -> Optional[DicomizerImageData]:

        if len(self._tiler.thumbnails) == 0:
            return None
        return OpenTileLevelImageData(
            self._tiler.thumbnails[0],
            self.base_metadata.pyramid.image,
            self.metadata.pyramid.image,
            self._encoder,
            self._volume_imaged_size,
            self._force_transcoding,
        )

    @cached_property
    def _volume_imaged_size(self):
        """Return the imaged size of the volume."""
        base_level = self._tiler.levels[0]
        return SizeMm(*base_level.pixel_spacing.to_tuple()) * Size(
            *base_level.image_size.to_tuple()
        )



#    Copyright 2023, 2025 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Image data for tiffslide compatible file."""


from functools import cached_property
from typing import Optional, Tuple, Union

import numpy as np
from PIL import Image as Pillow
from PIL.Image import Image
from tiffslide import TiffSlide
from wsidicom.codec import Encoder
from wsidicom.errors import WsiDicomNotFoundError
from wsidicom.geometry import Point, Region, Size
from wsidicom.metadata import Image as ImageMetadata

from wsidicomizer.sources.openslide_like import OpenSlideLikeLevelImageData


class TiffSlideLevelImageData(OpenSlideLikeLevelImageData):
    def __init__(
        self,
        tiff_slide: TiffSlide,
        blank_color: Optional[Union[int, Tuple[int, int, int]]],
        offset: Optional[Point],
        size: Optional[Size],
        image_metadata: ImageMetadata,
        level_index: int,
        tile_size: Optional[int],
        encoder: Encoder,
    ):
        """Wraps a TiffSlide level to ImageData.

        Parameters
        ----------
        tiff_slide: TiffSlide
            TiffSlide object to wrap.
        image_metadata: ImageMetadata
            Image metadata for image.
        level_index: int
            Level in TiffSlide object to wrap
        tile_size: int
            Output tile size.
        encoded: Encoder
            Encoder to use.
        """
        super().__init__(
            blank_color,
            offset,
            size,
            tiff_slide.level_dimensions,
            tiff_slide.level_downsamples,
            image_metadata,
            level_index,
            tile_size,
            encoder,
        )
        self._slide = tiff_slide

    @cached_property
    def samples_per_pixel(self) -> int:
        axes = self._slide.properties["tiffslide.series-axes"]
        if axes == "YX":
            return 1
        return 3

    def stitch_tiles(self, region: Region, path: str, z: float, threads: int) -> Image:
        """Overrides ImageData stitch_tiles() to read reagion directly from
        tiffslide object.

        Parameters
        ----------
        region: Region
             Pixel region to stitch to image
        path: str
            Optical path
        z: float
            Z coordinate

        Returns
        ----------
        Image
            Stitched image
        """
        if z not in self.focal_planes:
            raise WsiDicomNotFoundError(f"focal plane {z}", str(self))
        if path not in self.optical_paths:
            raise WsiDicomNotFoundError(f"optical path {path}", str(self))
        image_data = self._get_region(region)
        if image_data is None:
            return self._get_blank_decoded_frame(region.size)
        return Pillow.fromarray(image_data)

    def _get_region(self, region: Region) -> Optional[np.ndarray]:
        """Return Image read from region in tiffslide image. If image data for
        region is blank, None is returned. Transparent pixels are made into
        background color

        Parameters
        ----------
        region: Region
            Region to get image for.

        Returns
        ----------
        Optional[np.ndarray]
            Image data of region, or None if region is blank.
        """
        if region.size.width < 0 or region.size.height < 0:
            raise ValueError("Negative size not allowed")

        location_in_base_level = region.start * self._downsample + self._offset

        region_data = self._slide.read_region(
            location_in_base_level.to_tuple(),
            self._level_index,
            region.size.to_tuple(),
            as_array=True,
        )
        if self._detect_blank_tile(region_data):
            return None
        if self.samples_per_pixel == 1:
            region_data = region_data.squeeze(2)
        return region_data

    def _get_encoded_tile(self, tile_point: Point, z: float, path: str) -> bytes:
        """Return image bytes for tile. Transparency is removed and tile is
        encoded as jpeg.

        Parameters
        ----------
        tile_point: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        bytes
            Tile bytes.
        """
        if z not in self.focal_planes:
            raise WsiDicomNotFoundError(f"focal plane {z}", str(self))
        if path not in self.optical_paths:
            raise WsiDicomNotFoundError(f"optical path {path}", str(self))
        tile = self._get_region(Region(tile_point * self.tile_size, self.tile_size))
        if tile is None:
            return self._get_blank_encoded_frame(self.tile_size)
        return self.encoder.encode(tile)

    def _get_decoded_tile(self, tile_point: Point, z: float, path: str) -> Image:
        """Return Image for tile. Image mode is RGB.

        Parameters
        ----------
        tile_point: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        Image
            Tile as Image.
        """
        if z not in self.focal_planes:
            raise WsiDicomNotFoundError(f"focal plane {z}", str(self))
        if path not in self.optical_paths:
            raise WsiDicomNotFoundError(f"optical path {path}", str(self))
        tile = self._get_region(Region(tile_point * self.tile_size, self.tile_size))
        if tile is None:
            return self._get_blank_decoded_frame(self.tile_size)
        return Pillow.fromarray(tile)

    def _detect_blank_tile2(self, data: np.ndarray) -> bool:
        """Detect if tile data is a blank tile, i.e. either has full
        transparency or is filled with background color. First checks if the
        corners are transparent or has background color before checking whole
        data.

        Parameters
        ----------
        data: np.ndarray
            Data to check if blank.

        Returns
        ----------
        bool
            True if tile is blank.
        """

        TOP = RIGHT = -1
        BOTTOM = LEFT = 0
        CORNERS_Y = [BOTTOM, BOTTOM, TOP, TOP]
        CORNERS_X = [LEFT, RIGHT, LEFT, RIGHT]
        background = np.array(self.blank_color)
        corners_rgb = np.ix_(CORNERS_X, CORNERS_Y)
        if np.all(data[corners_rgb] == background):
            if np.all(data == background):
                return True
        return False



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
from enum import Enum
from pathlib import Path
from typing import Optional, Tuple

import click
from wsidicom.codec.settings import (
    HTJpeg2000Settings,
    Jpeg2kSettings,
    JpegSettings,
    JpegXlSettings,
    Subsampling,
)
from wsidicom.file import OffsetTableType
from wsidicom.metadata.schema.json.wsi import WsiMetadataJsonSchema
from wsidicom.metadata.wsi import WsiMetadata

from wsidicomizer.wsidicomizer import SourceIdentifier, WsiDicomizer


class CliEncodingsOptions(Enum):
    JPEG = "jpeg"
    JPEG2000 = "jpeg2000"
    HTJPEG2000 = "htjpeg2000"
    JPEGXL = "jpegxl"


@click.command()
@click.option(
    "-i",
    "--input",
    "input_path",
    type=click.Path(exists=True, path_type=Path),
    required=True,
    help="Path to input wsi file.",
)
@click.option(
    "-o",
    "--output",
    "output_path",
    type=click.Path(path_type=Path),
    help=(
        "Path to output folder. Folder will be created and must not "
        "exist. If not specified a folder named after the input file "
        "is created in the same path."
    ),
)
@click.option(
    "-t",
    "--tile-size",
    type=int,
    default=512,
    help=(
        "Tile size (same for width and height). Required for ndpi and "
        "openslide formats."
    ),
)
@click.option(
    "-m",
    "--metadata",
    type=click.Path(exists=True, path_type=Path),
    help=("Path to json metadata that will override metadata from source image file."),
)
@click.option(
    "-d",
    "--default-metadata",
    type=click.Path(exists=True, path_type=Path),
    help="Path to json metadata that will be used as default values.",
)
@click.option(
    "-l",
    "--levels",
    type=int,
    multiple=True,
    help=(
        "Pyramid levels to include, if not all. E.g. 0 1 for base and "
        "first pyramid layer. Can be specified multiple times."
    ),
)
@click.option(
    "--add-missing-levels",
    is_flag=True,
    help="If to add missing dyadic levels up to the single tile level.",
)
@click.option(
    "--label",
    type=click.Path(exists=True, path_type=Path),
    help="Optional label image to use instead of label found in file.",
)
@click.option("--no-label", is_flag=True, help="If not to include label")
@click.option("--no-overview", is_flag=True, help="If not to include overview")
@click.option(
    "--no-confidential", is_flag=True, help="If not to include confidential metadata"
)
@click.option(
    "-w",
    "--workers",
    type=int,
    default=os.cpu_count(),
    help="Number of worker threads to use",
)
@click.option(
    "--chunk-size",
    type=int,
    default=100,
    help="Number of tiles to give each worker at a time",
)
@click.option(
    "--format",
    "encoding_format",
    type=click.Choice(CliEncodingsOptions, case_sensitive=False),
    default=CliEncodingsOptions.JPEG,
    help="Encoding format to use if re-encoding.",
)
@click.option(
    "--quality",
    type=float,
    default=None,
    help=(
        "Quality to use if re-encoding. It is not recommended to use > 95 for "
        "jpeg. Use < 1 or > 1000 for lossless jpeg2000."
    ),
)
@click.option(
    "--subsampling",
    type=click.Choice(Subsampling, case_sensitive=False),
    default=None,
    help=(
        "Subsampling option if using jpeg for re-encoding. Use '444' "
        "for no subsampling, '422' for 2x1 subsampling, and '420' for "
        "2x2 subsampling."
    ),
)
@click.option(
    "--offset-table",
    type=click.Choice(
        [OffsetTableType.BASIC, OffsetTableType.EXTENDED, OffsetTableType.EMPTY],
        case_sensitive=False,
    ),
    default=OffsetTableType.BASIC,
    help=("Offset table to use."),
)
@click.option(
    "--source",
    type=click.Choice(SourceIdentifier, case_sensitive=False),
    default=None,
    help=(
        "Source library to use for reading the input file. If not specified, "
        "the library will be chosen based on file type."
    ),
)
def main(
    input_path: Path,
    output_path: Optional[Path],
    tile_size: int,
    metadata: Optional[Path],
    default_metadata: Optional[Path],
    levels: Tuple[int, ...],
    add_missing_levels: bool,
    label: Optional[Path],
    no_label: bool,
    no_overview: bool,
    no_confidential: bool,
    workers: int,
    chunk_size: int,
    encoding_format: CliEncodingsOptions,
    quality: Optional[float],
    subsampling: Optional[str],
    offset_table: OffsetTableType,
    source: Optional[SourceIdentifier] = None,
):
    """Convert compatible wsi file to DICOM. The cli only supports a subset of the functionality
    of the WsiDicomizer class. For more advanced usage, use the class directly."""

    # Load metadata if provided
    loaded_metadata = None
    if metadata:
        loaded_metadata = _load_metadata(metadata)

    loaded_default_metadata = None
    if default_metadata:
        loaded_default_metadata = _load_metadata(default_metadata)

    # Convert levels tuple to list or None
    include_levels = list(levels) if levels else None

    # Create encoding settings
    encoding_settings = None
    if encoding_format == CliEncodingsOptions.JPEG:
        subsampling_enum = (
            Subsampling.from_string(subsampling)
            if subsampling is not None
            else Subsampling.R420
        )
        encoding_settings = JpegSettings(
            quality=int(quality) if quality else 80, subsampling=subsampling_enum
        )
    elif encoding_format == CliEncodingsOptions.JPEG2000:
        encoding_settings = Jpeg2kSettings(levels=[int(quality) if quality else 80])
    elif encoding_format == CliEncodingsOptions.HTJPEG2000:
        encoding_settings = HTJpeg2000Settings(levels=[int(quality) if quality else 80])
    elif encoding_format == CliEncodingsOptions.JPEGXL:
        encoding_settings = JpegXlSettings(level=int(quality) if quality else 90)
    else:
        raise ValueError(f"Unsupported encoding format {encoding_format}")

    WsiDicomizer.convert(
        filepath=input_path,
        output_path=output_path,
        metadata=loaded_metadata,
        default_metadata=loaded_default_metadata,
        tile_size=tile_size,
        add_missing_levels=add_missing_levels,
        include_levels=include_levels,
        include_label=not no_label,
        include_overview=not no_overview,
        include_confidential=not no_confidential,
        workers=workers,
        chunk_size=chunk_size,
        encoding=encoding_settings,
        offset_table=offset_table,
        label=label,
        preferred_source=source,
    )


def _load_metadata(filepath: Path) -> WsiMetadata:
    with open(filepath) as json_file:
        metadata = WsiMetadataJsonSchema().loads(json_file.read())
        assert isinstance(metadata, WsiMetadata)
        return metadata


if __name__ == "__main__":
    main()



#    Copyright 2025 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
from PIL import Image as Pillow
from PIL.Image import Image
from pydicom.uid import UID
from wsidicom.codec import Encoder
from wsidicom.geometry import Point, Size, SizeMm
from wsidicom.metadata import Image as ImageMetadata
from wsidicom.metadata import ImageCoordinateSystem

from wsidicomizer.config import settings
from wsidicomizer.image_data import DicomizerImageData


class OpenSlideLikeImageData(DicomizerImageData):
    def __init__(
        self,
        blank_color: Optional[Union[int, Tuple[int, int, int]]],
        encoder: Encoder,
    ):
        super().__init__(encoder)
        if blank_color is None:
            blank_color = self._get_blank_color(encoder.photometric_interpretation)
        self._blank_color = blank_color

    @property
    def transfer_syntax(self) -> UID:
        return self.encoder.transfer_syntax

    @property
    def photometric_interpretation(self) -> str:
        return self.encoder.photometric_interpretation

    @property
    def samples_per_pixel(self) -> int:
        return 3

    @property
    def focal_planes(self) -> List[float]:
        return [0.0]

    @property
    def optical_paths(self) -> List[str]:
        return ["0"]

    @property
    def blank_color(self) -> Union[int, Tuple[int, int, int]]:
        return self._blank_color

    @property
    def thread_safe(self) -> bool:
        return True

    def _remove_alpha(self, alpha_image: Image) -> Image:
        """Remove alpha channel from image.

        Parameters
        ----------
        alpha_image: Image
            Image to remove alpha from.

        Returns
        ----------
        Image
            Image without alpha channel.
        """
        image = Pillow.new("RGB", alpha_image.size, self.blank_color)
        image.paste(alpha_image, mask=alpha_image.split()[3])
        return image


class OpenSlideLikeSingleImageData(OpenSlideLikeImageData):
    def __init__(
        self,
        image: Image,
        blank_color: Optional[Union[int, Tuple[int, int, int]]],
        encoder: Encoder,
    ) -> None:
        super().__init__(
            blank_color,
            encoder,
        )
        if image.mode == "RGBA":
            image = self._remove_alpha(image)
        self._image_size = Size.from_tuple(image.size)
        self._decoded_image = image
        self._encoded_image = self.encoder.encode(np.asarray(image))
        self._image_size = Size.from_tuple(image.size)

    @property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return self._image_size

    @property
    def tile_size(self) -> Size:
        """The pixel tile size of the image."""
        return self._image_size

    def _get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        if tile != Point(0, 0):
            raise ValueError("Point(0, 0) only valid tile for non-tiled image")
        return self._encoded_image

    def _get_decoded_tile(self, tile: Point, z: float, path: str) -> Image:
        if tile != Point(0, 0):
            raise ValueError("Point(0, 0) only valid tile for non-tiled image")
        return self._decoded_image


class OpenSlideLikeAssociatedImageData(OpenSlideLikeSingleImageData):
    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        """Size of the pixels in mm/pixel."""
        return None

    @property
    def imaged_size(self) -> Optional[Size]:
        return None


class OpenSlideLikeThumbnailImageData(OpenSlideLikeSingleImageData):
    def __init__(
        self,
        image: Image,
        blank_color: Optional[Union[int, Tuple[int, int, int]]],
        offset: Optional[Point],
        size: Optional[Size],
        level_dimensions: Sequence[Tuple[int, int]],
        image_metadata: ImageMetadata,
        encoder: Encoder,
    ) -> None:
        self._image_coordinate_system = image_metadata.image_coordinate_system
        if image_metadata.pixel_spacing is None:
            raise ValueError(
                "Could not determine pixel spacing for openslide thumbnail image."
            )
        base_level_dimensions = level_dimensions[0]
        downsample = (
            SizeMm.from_tuple(base_level_dimensions).width
            / Size.from_tuple(image.size).width
        )
        self._pixel_spacing = SizeMm(
            image_metadata.pixel_spacing.width * downsample,
            image_metadata.pixel_spacing.height * downsample,
        )

        if offset is not None:
            crop_offset = Point(
                round(offset.x / downsample), round(offset.y / downsample)
            )
        else:
            crop_offset = Point(0, 0)
        if size is not None:
            crop_size = Size(
                round(size.width / downsample), round(size.height / downsample)
            )
            self._imaged_size = image_metadata.pixel_spacing * size
        else:
            crop_size = Size.from_tuple(image.size) - crop_offset
            self._imaged_size = image_metadata.pixel_spacing * Size.from_tuple(
                base_level_dimensions
            )

        cropped = image.crop(
            (
                crop_offset.x,
                crop_offset.y,
                crop_offset.x + crop_size.width,
                crop_offset.y + crop_size.height,
            )
        )
        super().__init__(
            cropped,
            blank_color,
            encoder,
        )

    @property
    def image_coordinate_system(self) -> ImageCoordinateSystem:
        if self._image_coordinate_system is None:
            return super().image_coordinate_system
        return self._image_coordinate_system

    @property
    def pixel_spacing(self) -> SizeMm:
        """Size of the pixels in mm/pixel."""
        return self._pixel_spacing

    @property
    def imaged_size(self) -> SizeMm:
        return self._imaged_size


class OpenSlideLikeLevelImageData(OpenSlideLikeImageData):
    def __init__(
        self,
        blank_color: Optional[Union[int, Tuple[int, int, int]]],
        offset: Optional[Point],
        size: Optional[Size],
        level_dimensions: Sequence[Tuple[int, int]],
        level_downsamples: Sequence[float],
        image_metadata: ImageMetadata,
        level_index: int,
        tile_size: Optional[int],
        encoder: Encoder,
    ):
        super().__init__(
            blank_color,
            encoder,
        )
        if tile_size is None:
            tile_size = settings.default_tile_size
        self._tile_size = Size(tile_size, tile_size)
        self._level_index = level_index
        self._downsample = level_downsamples[self._level_index]
        if image_metadata.pixel_spacing is None:
            raise ValueError(
                "Could not determine pixel spacing for tiffslide level image."
            )
        self._pixel_spacing = SizeMm(
            image_metadata.pixel_spacing.width * self._downsample,
            image_metadata.pixel_spacing.height * self._downsample,
        )

        self._offset = offset or Point(0, 0)

        if size is not None:
            self._image_size = size // int(round(self._downsample))
            self._imaged_size = image_metadata.pixel_spacing * size
        else:
            self._image_size = Size.from_tuple(level_dimensions[self._level_index])
            self._imaged_size = image_metadata.pixel_spacing * Size.from_tuple(
                level_dimensions[0]
            )

        self._image_coordinate_system = image_metadata.image_coordinate_system

    @property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return self._image_size

    @property
    def tile_size(self) -> Size:
        """The pixel tile size of the image."""
        return self._tile_size

    @property
    def pixel_spacing(self) -> SizeMm:
        """Size of the pixels in mm/pixel."""
        return self._pixel_spacing

    @property
    def imaged_size(self) -> SizeMm:
        return self._imaged_size

    @property
    def image_coordinate_system(self) -> ImageCoordinateSystem:
        if self._image_coordinate_system is None:
            return super().image_coordinate_system
        return self._image_coordinate_system



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Metadata for opentile file."""

from typing import Optional

from opentile import Metadata
from wsidicom.metadata import Equipment, Image, Label, OpticalPath, Overview, Pyramid

from wsidicomizer.metadata import WsiDicomizerMetadata


class OpenTileMetadata(WsiDicomizerMetadata):
    def __init__(
        self,
        metadata: Metadata,
        has_label: bool,
        has_overview: bool,
        icc_profile: Optional[bytes] = None,
    ):
        equipment = Equipment(
            metadata.scanner_manufacturer,
            metadata.scanner_model,
            metadata.scanner_serial_number,
            metadata.scanner_software_versions,
        )
        image = Image(metadata.aquisition_datetime)
        if icc_profile is not None:
            optical_path = OpticalPath(icc_profile=icc_profile)
            optical_paths = [optical_path]
        else:
            optical_paths = []
        pyramid = Pyramid(image=image, optical_paths=optical_paths)
        if has_label:
            label = Label(image=Image(metadata.aquisition_datetime), optical_paths=[])
        else:
            label = None
        if has_overview:
            overview = Overview(
                image=Image(metadata.aquisition_datetime),
                optical_paths=[],
            )
        else:
            overview = None
        super().__init__(
            equipment=equipment, pyramid=pyramid, label=label, overview=overview
        )



#    Copyright 2023, 2025 SECTRA AB

#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Source for reading tiffslide compatible file."""


from pathlib import Path
from typing import Optional, Union

from pydicom import Dataset
from tiffslide import TiffSlide
from tiffslide.tiffslide import (
    PROPERTY_NAME_BACKGROUND_COLOR,
    PROPERTY_NAME_BOUNDS_HEIGHT,
    PROPERTY_NAME_BOUNDS_WIDTH,
    PROPERTY_NAME_BOUNDS_X,
    PROPERTY_NAME_BOUNDS_Y,
    PROPERTY_NAME_MPP_X,
    PROPERTY_NAME_MPP_Y,
    PROPERTY_NAME_OBJECTIVE_POWER,
    PROPERTY_NAME_VENDOR,
)
from wsidicom.codec import Encoder
from wsidicom.metadata import WsiMetadata

from wsidicomizer.image_data import DicomizerImageData
from wsidicomizer.metadata import MetadataPostProcessor
from wsidicomizer.sources.openslide_like import (
    OpenSlideLikeProperties,
    OpenSlideLikeSource,
)
from wsidicomizer.sources.openslide_like.openslide_like_metadata import (
    OpenSlideLikeMetadata,
)
from wsidicomizer.sources.tiffslide.tiffslide_image_data import (
    TiffSlideLevelImageData,
)


class TiffSlideSource(OpenSlideLikeSource):
    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: Optional[int] = None,
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        include_confidential: bool = True,
        metadata_post_processor: Optional[Union[Dataset, MetadataPostProcessor]] = None,
        **source_args,
    ) -> None:
        """Create a new TiffSlideSource.

        Parameters
        ----------
        filepath: Path
            Path to the file.
        encoder: Encoder
            Encoder to use. Pyramid is always re-encoded using the encoder.
        tile_size: Optional[int] = None,
            Tile size to use. If None, the default tile size is used.
        metadata: Optional[WsiMetadata] = None
            User-specified metadata that will overload metadata from source image file.
        default_metadata: Optional[WsiMetadata] = None
            User-specified metadata that will be used as default values.
        include_confidential: bool = True
            Include confidential metadata.
        metadata_post_processor: Optional[Union[Dataset, MetadataPostProcessor]] = None
            Optional metadata post processing by update from dataset or callback.

        """
        self._tiffslide = TiffSlide(filepath, **source_args)
        properties = OpenSlideLikeProperties(
            background_color=self._tiffslide.properties.get(
                PROPERTY_NAME_BACKGROUND_COLOR
            ),
            bounds_x=self._tiffslide.properties.get(PROPERTY_NAME_BOUNDS_X),
            bounds_y=self._tiffslide.properties.get(PROPERTY_NAME_BOUNDS_Y),
            bounds_height=self._tiffslide.properties.get(PROPERTY_NAME_BOUNDS_HEIGHT),
            bounds_width=self._tiffslide.properties.get(PROPERTY_NAME_BOUNDS_WIDTH),
            objective_power=self._tiffslide.properties.get(
                PROPERTY_NAME_OBJECTIVE_POWER
            ),
            vendor=self._tiffslide.properties.get(PROPERTY_NAME_VENDOR),
            mpp_x=self._tiffslide.properties.get(PROPERTY_NAME_MPP_X),
            mpp_y=self._tiffslide.properties.get(PROPERTY_NAME_MPP_Y),
        )

        super().__init__(
            filepath,
            properties,
            self._tiffslide.level_downsamples,
            self._tiffslide.level_dimensions,
            self._tiffslide.associated_images,
            OpenSlideLikeMetadata(properties, self._tiffslide.color_profile),
            encoder,
            tile_size,
            metadata,
            default_metadata,
            include_confidential,
            metadata_post_processor,
        )

    def close(self):
        self._tiffslide.close()

    @staticmethod
    def is_supported(filepath: Path) -> bool:
        """Return True if file in filepath is supported by TiffSlide."""
        format = TiffSlide.detect_format(filepath)
        return format is not None

    def _create_level_image_data(
        self, level_index: int
    ) -> Optional[DicomizerImageData]:
        return TiffSlideLevelImageData(
            self._tiffslide,
            self._blank_color,
            self._base_level_offset,
            self._base_level_size,
            self.metadata.pyramid.image,
            level_index,
            self._tile_size,
            self._encoder,
        )



#    Copyright 2025 SECTRA AB

#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Source for reading openslide like compatible file."""

import math
import re
from enum import Enum
from pathlib import Path
from typing import Dict, Mapping, Optional, Sequence, Tuple, Union

from PIL.Image import Image
from pydicom import Dataset
from wsidicom.codec import Encoder
from wsidicom.geometry import Point, Size
from wsidicom.metadata import WsiMetadata

from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.image_data import DicomizerImageData
from wsidicomizer.metadata import MetadataPostProcessor
from wsidicomizer.sources.openslide_like import (
    OpenSlideLikeAssociatedImageData,
    OpenSlideLikeProperties,
    OpenSlideLikeThumbnailImageData,
)
from wsidicomizer.sources.openslide_like.openslide_like_metadata import (
    OpenSlideLikeMetadata,
)


class OpenSlideLikeAssociatedImageType(Enum):
    LABEL = "label"
    MACRO = "macro"
    THUMBNAIL = "thumbnail"


class OpenSlideLikeSource(DicomizerSource):
    def __init__(
        self,
        filepath: Path,
        properties: OpenSlideLikeProperties,
        level_downsamples: Sequence[float],
        level_dimensions: Sequence[Tuple[int, int]],
        associated_images: Mapping[str, Image],
        base_metadata: OpenSlideLikeMetadata,
        encoder: Encoder,
        tile_size: Optional[int] = None,
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        include_confidential: bool = True,
        metadata_post_processor: Optional[Union[Dataset, MetadataPostProcessor]] = None,
    ) -> None:
        """Create a new OpenSlideLikeSource.

        Parameters
        ----------
        filepath: Path
            Path to the file.
        properties: OpenSlideLikeProperties
            Properties of the file.
        level_downsamples: Sequence[float]
            Downsamples of the levels.
        level_dimensions: Sequence[Tuple[int, int]]
            Dimensions of the levels.
        associated_images: Mapping[str, Image]
            Associated images in the file.
        base_metadata: OpenSlideLikeMetadata
            Base metadata for the file.
        encoder: Encoder
            Encoder to use. Pyramid is always re-encoded using the encoder.
        tile_size: Optional[int] = None,
            Tile size to use. If None, the default tile size is used.
        metadata: Optional[WsiMetadata] = None
            User-specified metadata that will overload metadata from source image file.
        default_metadata: Optional[WsiMetadata] = None
            User-specified metadata that will be used as default values.
        include_confidential: bool = True
            Include confidential metadata.
        metadata_post_processor: Optional[Union[Dataset, MetadataPostProcessor]] = None
            Optional metadata post processing by update from dataset or callback.

        """
        self._base_metadata = base_metadata
        self._level_downsamples = level_downsamples
        self._level_dimensions = level_dimensions
        self._associated_images = associated_images
        self._pyramid_levels = {
            (int(round(math.log2(downsample))), 0.0, "0"): index
            for index, downsample in enumerate(level_downsamples)
        }

        self._blank_color = self._get_blank_color(properties)
        self._base_level_offset, self._base_level_size = self._get_offset_and_size(
            properties
        )
        super().__init__(
            filepath,
            encoder,
            tile_size,
            metadata,
            default_metadata,
            include_confidential,
            metadata_post_processor,
        )

    @property
    def base_metadata(self) -> WsiMetadata:
        return self._base_metadata

    @property
    def pyramid_levels(self) -> Dict[Tuple[int, float, str], int]:
        return self._pyramid_levels

    def _create_label_image_data(self) -> Optional[DicomizerImageData]:
        label_image = self._get_associated_image(OpenSlideLikeAssociatedImageType.LABEL)
        if label_image is None:
            return None
        return OpenSlideLikeAssociatedImageData(
            label_image,
            self._blank_color,
            self._encoder,
        )

    def _create_overview_image_data(self) -> Optional[DicomizerImageData]:
        overview_image = self._get_associated_image(
            OpenSlideLikeAssociatedImageType.MACRO
        )
        if overview_image is None:
            return None
        return OpenSlideLikeAssociatedImageData(
            overview_image,
            self._blank_color,
            self._encoder,
        )

    def _create_thumbnail_image_data(self) -> Optional[DicomizerImageData]:
        label_image = self._get_associated_image(
            OpenSlideLikeAssociatedImageType.THUMBNAIL
        )
        if label_image is None:
            return None
        return OpenSlideLikeThumbnailImageData(
            label_image,
            self._blank_color,
            self._base_level_offset,
            self._base_level_size,
            self._level_dimensions,
            self.metadata.pyramid.image,
            self._encoder,
        )

    def _get_associated_image(
        self, image_type: OpenSlideLikeAssociatedImageType
    ) -> Optional[Image]:
        """Get image from associated images.

        Parameters
        ----------
        associated_images: Mapping[str, Image]
            Associated images to get image from.
        image_type: str
            Type of image to get.

        Returns
        ----------
        Image
            Associated image.
        """
        return self._associated_images.get(image_type.value)

    @staticmethod
    def _get_blank_color(
        properties: OpenSlideLikeProperties,
    ) -> Optional[Union[int, Tuple[int, int, int]]]:
        if properties.background_color is not None:
            rgb = re.findall(r"([0-9a-fA-F]{2})", properties.background_color)
            if len(rgb) == 3:
                return (int(rgb[0], 16), int(rgb[1], 16), int(rgb[2], 16))
        return None

    @staticmethod
    def _get_offset_and_size(
        properties: OpenSlideLikeProperties,
    ) -> Tuple[Optional[Point], Optional[Size]]:
        if properties.bounds_x is not None and properties.bounds_y is not None:
            offset = Point(int(properties.bounds_x), int(properties.bounds_y))
        else:
            offset = None
        if properties.bounds_width is not None and properties.bounds_height is not None:
            size = Size(int(properties.bounds_width), int(properties.bounds_height))
        else:
            size = None
        return offset, size
